#include <linux/module.h> // MODULE_*
#include <linux/init.h> // __init, __exit
#include <linux/string.h> // sprintf, sscanf
#include <linux/uaccess.h> // copy_from_user, copy_to_user
#include <linux/printk.h> // pr_info
#include <linux/sysfs.h> // sysfs_*
#include <linux/kobject.h> // kobj_*
#include <linux/proc_fs.h> // proc_*

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Pablo");
MODULE_DESCRIPTION("Provides a /proc/accum file that accumulates the sum of what it is written.");
MODULE_VERSION("0.1");

/* accumulator state */

static long sum = 0;
static char sum_buf[32] = "0\n";
static int sum_buf_len = 2;

/* multiplier parameter */

static struct kobject *accum_obj;
static int multiplier = 1;

static ssize_t mult_show (struct kobject *kobj, struct kobj_attribute *attr, char *buf)
{
  return sprintf(buf, "%d\n", multiplier);
}

static ssize_t mult_store (struct kobject *kobj, struct kobj_attribute *attr, char *buf, size_t len)
{
  sscanf(buf, "%d", &multiplier);
  pr_info("accum: multiplier is now %d\n", multiplier);
  return len;
}

static struct kobj_attribute mult_attr = __ATTR(multiplier, 0644, mult_show, (void *)mult_store);

/* /proc entry */

static struct proc_dir_entry *proc_accum;

static ssize_t
accum_read (struct file *f, char __user *buf, size_t len, loff_t *offset)
{
  int l = ((sum_buf_len - *offset) > len) ? len : (sum_buf_len - *offset);
  if (l <= 0) return 0;
  if (copy_to_user(buf, sum_buf + *offset, l) > 0) {
    pr_info("accum: failed to copy\n");
    return 0;
  }
  *offset += l;
  return l;
}

static ssize_t
accum_write (struct file *f, const char __user *buf, size_t len, loff_t *offset)
{
  char tmp[32];
  int l = (len < 32) ? len : 31;
  int n;
  if (copy_from_user(tmp, buf, l) > 0) {
    pr_info("accum: failed to copy\n");
    return -1;
  }
  sscanf(tmp, "%d", &n);
  sum += n * multiplier;
  sum_buf_len = sprintf(sum_buf, "%ld\n", sum);
  *offset = 0;
  pr_info("accum: added %d\n", n * multiplier);
  return l;
}

static const struct proc_ops accum_fops = {
  .proc_read = accum_read,
  .proc_write = accum_write,
};

/* init and cleanup */

static int __init accum_init (void)
{
  printk(KERN_INFO "accum: module initializes\n");

  if ((accum_obj = kobject_create_and_add("accum", kernel_kobj)) == NULL) {
    pr_alert("accum: error: could not create kobject accum\n");
    return -1;
  }
  if (sysfs_create_file(accum_obj, &mult_attr.attr) != 0) {
    pr_alert("accum: error: could not create multiplier file in /sys/kernel/accum\n");
    kobject_put(accum_obj);
    return -1;
  }
  pr_info("accum: /sys/kernel/accum/multiplier created\n");

  if ((proc_accum = proc_create("accum", 0666, NULL, &accum_fops)) == NULL) {
    pr_alert("accum: error: could not create /proc/accum\n");
    return -1;
  }
  pr_info("accum: /proc/accum created\n");

  return 0;
}

static void __exit accum_exit (void)
{
  proc_remove(proc_accum);
  kobject_put(accum_obj);
  printk(KERN_INFO "accum: module exits\n");
}

module_init(accum_init);
module_exit(accum_exit);
